/* 
 * Adium is the legal property of its developers, whose names are listed in the copyright file included
 * with this source distribution.
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
 * the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; if not,
 * write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#import <Adium/AIAbstractAccount.h>
#import <Adium/AIAccount.h>
#import <Adium/AIListContact.h>
#import <Adium/AIContentMessage.h>
#import <Adium/AIContentNotification.h>
#import <Adium/AIService.h>
#import <Adium/AIChat.h>
#import <Adium/ESFileTransfer.h>
#import "AdiumAccounts.h"

#import <AIChatCommandsController.h>
#import <Adium/AIContactControllerProtocol.h>
#import <Adium/AIContentControllerProtocol.h>
#import <Adium/AIAccountControllerProtocol.h>

#import "DCJoinChatViewController.h"
#import "AIChatControllerProtocol.h"
#import "AIMessageWindowController.h"
#import "AIMessageWindow.h"
#import "AIInterfaceControllerProtocol.h"
#import "AIStatusControllerProtocol.h"

#define NEW_ACCOUNT_DISPLAY_TEXT			AILocalizedString(@"<New Account>", "Placeholder displayed as the name of a new account")

@interface AIAccountDeletionDialog : NSObject <AIAccountControllerRemoveConfirmationDialog> {
	AIAccount *account;
	NSAlert *alert;
	id userdata;
}

- (id)initWithAccount:(AIAccount*)ac alert:(NSAlert*)al;

- (void)setUserData:(id)ud;
- (id)userData;

- (void)alertDidEnd:(NSAlert *)alert returnCode:(int)returnCode contextInfo:(void *)contextInfo;

@end

@implementation AIAccountDeletionDialog

- (id)initWithAccount:(AIAccount*)ac alert:(NSAlert*)al {
	if((self = [super init])) {
		account = ac;
		alert = [al retain];
	}
	return self;
}

- (void)dealloc {
	[alert release];
	[userdata release];
	[super dealloc];
}

- (void)setUserData:(id)ud {
	id old = userdata;
	userdata = [ud retain];
	[old release];
}

- (id)userData {
	return [[userdata retain] autorelease];
}

- (void)runModal {
	[self alertDidEnd:alert returnCode:[alert runModal] contextInfo:NULL];
}

- (void)beginSheetModalForWindow:(NSWindow*)window {
	[alert beginSheetModalForWindow:window modalDelegate:self didEndSelector:@selector(alertDidEnd:returnCode:contextInfo:) contextInfo:NULL];
}

- (void)alertDidEnd:(NSAlert *)alert returnCode:(int)returnCode contextInfo:(void *)contextInfo {
	[account alertForAccountDeletion:self didReturn:returnCode];
}

@end

/*!
 * @class AIAccount
 * @brief An account
 *
 * This abstract class represents an account the user has setup in Adium.  Subclass this for every service.
 */
@implementation AIAccount

/*!
 * @brief Init Account
 *
 * Init this account instance
 */
- (void)initAccount
{
}

/*!
 * @brief Connect
 *
 * Connect the account, transitioning it into an online state.
 */
- (void)connect
{
	//We are connecting
	[self setStatusObject:[NSNumber numberWithBool:YES] forKey:@"Connecting" notify:NotifyNow];
}

/*!
 * @brief rejoinChat
 * 
 * Rejoin the open group chats after disconnect
 */
-(BOOL)rejoinChat:(AIChat*)chat
{
	return NO;
}	

/*!
 * @brief Disconnect
 *
 * Disconnect the account, transitioning it into an offline state.
 */
- (void)disconnect
{
	[self cancelAutoReconnect];
	[self setStatusObject:nil forKey:@"Connecting" notify:NotifyLater];

	[self notifyOfChangedStatusSilently:NO];
}

/*!
 * @brief Disconnect as a result of the network connection dropping out
 *
 * The default implementation is identical to [self disconect], but subclasses may want to act differently
 * if the network connection is gone than if the user chose to disconnect.
 *
 * Subclasses should call super's implementation.
 */
- (void)disconnectFromDroppedNetworkConnection
{
	[self disconnect];
}

/*!
 * @brief Register an account
 *
 * Register an account on this service using the currently entered information.  This is for services which support
 * in-client registration such as jabber.
 */
- (void)performRegisterWithPassword:(NSString *)inPassword
{

}

/*!
 * @brief The UID will be changed. The account has a chance to perform modifications
 *
 * For example, MSN adds @hotmail.com to the proposedUID and returns the new value
 *
 * @param proposedUID The proposed, pre-filtered UID (filtered means it has no characters invalid for this servce)
 * @result The UID to use; the default implementation just returns proposedUID.
 */
- (NSString *)accountWillSetUID:(NSString *)proposedUID
{
	return proposedUID;
}

/*!
 * @brief The account's UID changed
 */
- (void)didChangeUID
{

}

/*!
 * @brief The account will be deleted
 *
 * The default implementation disconnects the account.  Subclasses should call super's implementation.
 * If asynchronous behavior is required, the next three methods should be overridden instead.
 */
- (void)willBeDeleted
{
	[self setShouldBeOnline:NO];

	//Remove our contacts immediately.
	[self removeAllContacts];
}

- (id<AIAccountControllerRemoveConfirmationDialog>)confirmationDialogForAccountDeletionForAccountsList:(AdiumAccounts*)accounts
{
	AIAccountDeletionDialog *result = [[AIAccountDeletionDialog alloc] initWithAccount:self alert:[self alertForAccountDeletion]];
	[result setUserData:accounts];
	return result;
}

/*!
 * @brief The alert used for confirming the account deletion
 *
 * Meant for subclassers. By default, returns the dialog that asks the user if the account should really be deleted (and how).
 */
- (NSAlert*)alertForAccountDeletion
{
	return [NSAlert alertWithMessageText:AILocalizedString(@"Delete Account",nil)
						   defaultButton:AILocalizedString(@"Delete",nil)
						 alternateButton:AILocalizedString(@"Cancel",nil)
							 otherButton:nil
			   informativeTextWithFormat:AILocalizedString(@"Delete the account %@?",nil), ([[self formattedUID] length] ? [self formattedUID] : NEW_ACCOUNT_DISPLAY_TEXT)];
}

/*!
 * @brief The dialog asking for confirmation for deleting the account did return.
 *
 * @parameter dialog The dialog that has completed
 * @parameter returnCode One of the regular NSAlert return codes
 *
 * This method should be overridden when alertForAccountDeletion: was overridden, and/or asynchronous behavior is required.
 * This implementation disconnects and deletes the account from the accounts list when returnCode == NSAlertDefaultReturn.
 * It must be called by subclassers (could be done asynchronously) with either NSAlertDefaultReturn or NSAlertAlternateReturn.
 */
- (void)alertForAccountDeletion:(id<AIAccountControllerRemoveConfirmationDialog>)dialog didReturn:(int)returnCode
{
	if(returnCode == NSAlertDefaultReturn) {
		[self willBeDeleted];
		[(AdiumAccounts*)[(AIAccountDeletionDialog*)dialog userData] deleteAccount:self];
	}
	[(AIAccountDeletionDialog*)dialog release];
}

/*!
 * @brief A formatted UID which may include additional necessary identifying information.
 *
 * For example, an AIM account (tekjew) and a .Mac account (tekjew@mac.com, entered only as tekjew) may appear identical
 * without service information (tekjew). The explicit formatted UID is therefore tekjew@mac.com
 */
- (NSString *)explicitFormattedUID
{
	return [self formattedUID];
}

//Properties -----------------------------------------------------------------------------------------------------------
#pragma mark Properties
/*!
 * @brief Send Autoresponses while away
 *
 * Subclass to alter the behavior of this account with regards to autoresponses.  Certain services expect the client to
 * auto-respond with away messages.  Adium will provide this behavior automatically if desired.
 */
- (BOOL)supportsAutoReplies
{
	return NO;
}

/*!
 * @brief Disconnect on fast user switch
 *
 * It may be required for a service to disconnect when logged in users change.  If this is the case, subclass this
 * method to return YES and Adium will automatically disconnect and reconnect on FUS events.
 */
- (BOOL)disconnectOnFastUserSwitch
{
	return NO;
}

/*!
 * @brief Connectivity based on network reachability
 *
 * By default, accounts are automatically disconnected and reconnected when network reachability changes.  Accounts
 * that do not require persistent network connections can choose to disable this by returning NO from this method.
 */
- (BOOL)connectivityBasedOnNetworkReachability
{
	return YES;
}

/*!
 * @brief Suppress typing notifications after send
 *
 * Some protocols require a 'Stopped typing' notification to be sent along with an instant message.  Other protocols
 * implicitly assume that typing has stopped with an incoming message and the extraneous typing notification may cause
 * strange behavior.  Return YES from this method to suppress the sending of a stopped typing notification along with
 * messages.
 */
- (BOOL)suppressTypingNotificationChangesAfterSend
{
	return NO;
}

/*!
 * @brief Support server-side storing of messages to offline users
 *
 * Some protocols store messages to offline contacts on the server. Subclasses may return YES if their service supports 
 * this. Adium will not store the message as an Event, and will just send it along to the server. This may cause a Gaim
 * error on Jabber if the Jabber server they are using is down.
 */
- (BOOL)canSendOfflineMessageToContact:(AIListContact *)inContact
{
	return NO;
}

/*!
 * @brief Are newlines allowed in messages?
 *
 * If NO, messages with newlines will be split into multiple messages before being sent
 */
- (BOOL)allowsNewlinesInMessages
{
	return YES;
}

/*!
 * @brief Does the account itself display file transfer messages in chat windows?
 *
 * If YES, Adium won't attempt to display messages in chat windows regarding file transfers.
 * If NO, Adium automatically displays appropriate messages in open chats.
 */
- (BOOL)accountDisplaysFileTransferMessages
{
	return NO;
}

/*!
 * @brief Called once the display name has been properly filtered
 *
 * Subclasses may override to pass this name on to the server if appropriate.
 * Super's implementation should then be called.
 */
- (void)gotFilteredDisplayName:(NSAttributedString *)attributedDisplayName
{
	[self updateLocalDisplayNameTo:attributedDisplayName];
}

- (NSImage *)userIcon
{
	NSData	*iconData = [self userIconData];
	return (iconData ? [[[NSImage alloc] initWithData:iconData] autorelease] : nil);
}

//Status ---------------------------------------------------------------------------------------------------------------
#pragma mark Status
/*!
 * @brief Supported status keys
 *
 * Returns an array of status keys supported by this account.  This account will not be informed of changes to keys
 * it does not support.  Available keys are:
 *   @"Display Name", @"Online", @"Offline", @"IdleSince", @"IdleManuallySet", @"User Icon"
 *   @"TextProfile", @"DefaultUserIconFilename", @"StatusState"
 * @return NSSet of supported keys
 */
- (NSSet *)supportedPropertyKeys
{
	static	NSSet	*supportedPropertyKeys = nil;
	if (!supportedPropertyKeys) {
		supportedPropertyKeys = [[NSSet alloc] initWithObjects:
			@"Online",
			@"FormattedUID",
			KEY_ACCOUNT_DISPLAY_NAME,
			@"Display Name",
			@"StatusState",
			KEY_USER_ICON,
			@"Enabled",
			nil];
	}

	return supportedPropertyKeys;
}

/*!
 * @brief Status for key
 *
 * Returns the status this account should be for a specific key
 * @param key Status key
 * @return id Status value
 */
- (id)statusForKey:(NSString *)key
{
	return [self preferenceForKey:key group:GROUP_ACCOUNT_STATUS];
}

/*!
 * @brief Update account status
 *
 * Update account status for the changed key.  This is called when account status changes Adium-side and the account
 * code should update status account/server side in response.  The new value for the key can be accessed using
 * the statusForKey method.
 * @param key The updated status key
 */
- (void)updateStatusForKey:(NSString *)key
{
	[self updateCommonStatusForKey:key];
}

/*!
 * @brief Update contact status
 *
 * Adium is requesting that the account update a contact's status.  This method is primarily called by the get info
 * window.  Since this is called sparsely, accounts may choose to look up additional information such as profiles
 * in response to this.  Adium guards this method to prevent it from being called too rapidly, so expensive lookups
 * are not a problem if the delayedUpdateStatusInterval is set correctly.
 */
- (void)delayedUpdateContactStatus:(AIListContact *)inContact
{
	
}

/*!
 * @brief Update contact interval
 *
 * Specifies the mininum interval at which delayedUpdateContactStatus will be called.  If the account code is performing
 * expensive operations (such as profile or web lookups) in response to updateContactStatus, it can guard against
 * the lookups being performed too frequently by returning an interval here.
 */
- (float)delayedUpdateStatusInterval
{
	return 3.0;
}

/*!
 * @brief Perform the setting of a status state
 *
 * Sets the account to a passed status state.  The account should set itself to best possible status given the return
 * values of statusState's accessors.  The passed statusMessage has been filtered; it should be used rather than
 * [statusState statusMessage], which returns an unfiltered statusMessage.
 *
 * @param statusState The state to enter
 * @param statusMessage The filtered status message to use.
 */
- (void)setStatusState:(AIStatus *)statusState usingStatusMessage:(NSAttributedString *)statusMessage
{
	
}

/*!
 * @brief Should the autorefreshing attributed string associated with a key be updated at the moment?
 *
 * The default implementation causes all dynamic strings which need updating to be updated if the account is
 * online.  Subclasses may choose to implement more complex logic; for example, a nickname seen only in a chat
 * might be updated only if a chat is open.
 */
- (BOOL)shouldUpdateAutorefreshingAttributedStringForKey:(NSString *)inKey
{
	return [self online];
}

//Messaging, Chatting, Strings -----------------------------------------------------------------------------------------
#pragma mark Messaging, Chatting, Strings
/*!
 * @brief Available for sending content
 *
 * Returns YES if the contact is available for receiving content of the specified type.  If contact is nil, instead
 * check for the availiability to send any content of the given type.
 *
 * The default implementation indicates the account, if online, can send messages to any online contact.
 * It can also send files to any online contact if the account subclass conforms to the AIAccount_Files protocol.
 *
 * @param inType A string content type
 * @param inContact The destination contact, or nil to check global availability
 */
- (BOOL)availableForSendingContentType:(NSString *)inType toContact:(AIListContact *)inContact
{
	if ([self online] && (!inContact || [inContact online])) {
		if ([inType isEqualToString:CONTENT_MESSAGE_TYPE] ||
			[inType isEqualToString:CONTENT_NOTIFICATION_TYPE]) {
			return YES;

		} else if ([inType isEqualToString:CONTENT_FILE_TRANSFER_TYPE]) {
			return [self conformsToProtocol:@protocol(AIAccount_Files)];
		}
	}

	return NO;
}

/*!
 * @brief Open a chat
 *
 * Open the passed chat account-side.  Depending on the protocol, account code may need to establish a connection in
 * response to this method or perhaps make no actions at all.  This method is used by both one-on-one chats and
 * multi-user chats.
 * @param chat The chat to open
 * @return YES on success
 */
- (BOOL)openChat:(AIChat *)chat
{
	return NO;
}

/*!
 * @brief Close a chat
 *
 * Close the passed chat account-side.  Depending on the protocol, account code may need to close a connection in
 * response to this method or perhaps make no actions at all.  This method is used by both one-on-one chats and
 * multi-user chats.
 *
 * This method should *only* be called by a core controller.  Call [[adium interfaceController] closeChat:chat] to perform a close from other code.
 *
 * @param chat The chat to close
 * @return YES on success
 */
- (BOOL)closeChat:(AIChat *)chat
{
	return NO;
}

/*!
 * @brief Invite a contact to an open chat
 *
 * Invite a contact to the passed chat, if supported by the protocol and the specific chat instance.  An invite
 * message is provided as a convenience to protocols that require or support one.
 * @param contact AIListObject to invite
 * @param chat AIChat they are being invited to
 * @param inviteMessage NSString invite message for the invited contact
 * @return YES on success
 */
- (BOOL)inviteContact:(AIListObject *)contact toChat:(AIChat *)chat withMessage:(NSString *)inviteMessage
{
	NSLog(@"invite contact to chat with message");
	return NO;
}

/*!
 * @brief Join a group chat
 *
 * Join a group chat by name
 * @param contact AIListObject to invite
 * @param chat AIChat they are being invited to
 * @param inviteMessage NSString invite message for the invited contact
 * @return YES on success
 */
- (BOOL)joinGroupChatNamed:(NSString *)name
{
	// XXX - Do we need this method?  All chats are supposed to be treated equally and assuming a 'name' seems protocol specific -ai
	return NO;
}

/*!
 * @brief Send a typing object
 *
 * The content object contains all the necessary information for sending,
 * including the destination contact.
 */
- (void)sendTypingObject:(AIContentTyping *)inTypingObject
{

}

/*!
 * @brief Send a message
 *
 * The content object contains all the necessary information for sending,
 * including the destination contact. [inMessageObject encodedMessage] contains the NSString which should be sent.
 */
- (BOOL)sendMessageObject:(AIContentMessage *)inMessageObject
{
	return NO;
}

/*!
 * @brief Encode attributed string (generic)
 *
 * Encode an NSAttributedString into a NSString for this account.  Accounts that support formatted text or require
 * special encoding on strings should do that work here.  For example, HTML based accounts should convert the 
 * NSAttributedString to HTML appropriate for their protocol (Adium can help with this).
 * @param inAttributedString String to encode
 * @param inListObject List object associated with the string; nil if the string is not associated with a particular list object, which is the case if encoding for a status message or a group chat message.
 * @return NSString result from encoding
 */
- (NSString *)encodedAttributedString:(NSAttributedString *)inAttributedString forListObject:(AIListObject *)inListObject
{
    return [inAttributedString string];
}

/*!
 * @brief Encode attributed string to send as a message
 */
- (NSString *)encodedAttributedStringForSendingContentMessage:(AIContentMessage *)inContentMessage
{
    return [self encodedAttributedString:[inContentMessage message] forListObject:[inContentMessage destination]];
}

//Presence Tracking ----------------------------------------------------------------------------------------------------
#pragma mark Presence Tracking
/*!
 * @brief Contact list editable?
 *
 * Returns YES if the contact list is currently editable
 * @param object AIContentObject to send
 * @return YES on success
 */
- (BOOL)contactListEditable
{
	return NO;
}

/*!
 * @brief Add contacts
 *
 * Add contacts to a group on this account.  Create the group if it doesn't already exist.
 * @param objects NSArray of AIListContact objects to add
 * @param group AIListGroup destination for contacts
 */
- (void)addContacts:(NSArray *)objects toGroup:(AIListGroup *)group
{
	//XXX - Our behavior for duplicate contacts isn't specified here.  Should we handle that adium-side automatically? -ai
}

/*!
 * @brief Remove contacts
 *
 * Remove contacts from this account.
 * @param objects NSArray of AIListContact objects to remove
 */
- (void)removeContacts:(NSArray *)objects
{
	
}

/*!
 * @brief Remove a group
 *
 * Remove a group from this account.
 * @param group AIListGroup to remove
 */
- (void)deleteGroup:(AIListGroup *)group
{
	//XXX - Adium's current behavior is to delete all the contacts within a group, and then delete the group.  This is innefficient on protocols which support deleting groups. -ai
}

/*!
 * @brief Move contacts
 *
 * Move existing contacts to a specific group on this account.  The passed contacts should already exist somewhere on
 * this account.
 * @param objects NSArray of AIListContact objects to remove
 * @param group AIListGroup destination for contacts
 */
- (void)moveListObjects:(NSArray *)objects toGroup:(AIListGroup *)group
{
	
}

/*!
 * @brief Rename a group
 *
 * Rename a group on this account.
 * @param group AIListGroup to rename
 * @param newName NSString name for the group
 */
- (void)renameGroup:(AIListGroup *)group to:(NSString *)newName
{
	
}

/*!
 * @brief Menu items for contact
 *
 * Returns an array of menu items for a contact on this account.  This is the best place to add protocol-specific
 * actions that aren't otherwise supported by Adium.
 * @param inContact AIListContact for menu items
 * @return NSArray of NSMenuItem instances for the passed contact
 */
- (NSArray *)menuItemsForContact:(AIListContact *)inContact
{
	return nil;
}

/*!
 * @brief Menu items for the account's actions
 *
 * Returns an array of menu items for account-specific actions.  This is the best place to add protocol-specific
 * actions that aren't otherwise supported by Adium.  It will only be queried if the account is online.
 * @return NSArray of NSMenuItem instances for this account
 */
- (NSArray *)accountActionMenuItems
{
	return nil;
}

/*!
 * @brief The account menu item was updated
 *
 * This method allows the opportunity to update the account menu item, e.g. to add information to it
 */
- (void)accountMenuDidUpdate:(NSMenuItem*)menuItem
{

}

/*!
 * @brief Is a contact on the contact list intentionally listed?
 *
 * By default, it is assumed that any contact on the list is intended be there.
 * This is used by AIListContact to determine if the prescence of itself on the list is indicative of a degree
 * of trust, for preferences such as "automatically accept files from contacts on my contact list".
 */
- (BOOL)isContactIntentionallyListed:(AIListContact *)contact
{
	return YES;
}

#pragma mark Secure messsaging

/*!
 * @brief Allow secure messaging toggling on a chat?
 *
 * Returns YES if secure (encrypted) messaging's status for this chat should be able to be changed.
 * This allows the account to determine on a per-chat basis whether the chat's initial security setting should be permanently
 * maintained.  If it returns NO, the user can not request for the chat to become encrypted or unencrypted.
 * This is currently implemented by Gaim accounts to return YES for one-on-one chats and NO for group chats to indicate
 * the functionality provided by Off-the-Record Messaging (OTR).
 *
 * @param inChat The query chat 
 * @result Should the state of secure messaging be allowed to change?
 */
- (BOOL)allowSecureMessagingTogglingForChat:(AIChat *)inChat
{
	//Allow secure messaging via OTR for one-on-one chats
	return ![inChat isGroupChat];
}

/*!
 * @brief Provide a localized description of the encryption this account provides
 *
 * Returns a localized string which describes the encryption this account supports.
 *
 * @result An <tt>NSString</tt> describing the encryption offerred by this account, if any.
 */
- (NSString *)aboutEncryption
{
	return [NSString stringWithFormat:
		AILocalizedStringFromTableInBundle(@"Adium provides encryption, authentication, deniability, and perfect forward secrecy over %@ via Off-the-Record Messaging (OTR). If your contact is not using an OTR-compatible messaging system, your contact will be sent a link to the OTR web site when you attempt to connect. For more information on OTR, visit http://www.cypherpunks.ca/otr/.", nil, [NSBundle bundleForClass:[AIAccount class]], nil),
		[[self service] shortDescription]];
}

/*!
 * @brief Start or stop secure messaging in a chat
 *
 * @param inSecureMessaging The desired state of the chat in terms of encryption
 * @param inChat The chat to change
 */
- (void)requestSecureMessaging:(BOOL)inSecureMessaging
						inChat:(AIChat *)inChat
{
	[[adium contentController] requestSecureOTRMessaging:inSecureMessaging
												  inChat:inChat];
}

/*!
 * @brief Allow the user to verify (or unverify) the identity being used for encryption in a chat
 *
 * It is an error to call this on a chat which is not current encrypted.
 *
 * @param The chat
 */
- (void)promptToVerifyEncryptionIdentityInChat:(AIChat *)inChat
{
	[[adium contentController] promptToVerifyEncryptionIdentityInChat:inChat];
}

#pragma mark Image sending
/*!
 * @brief Can the account send images inline within a chat?
 */
- (BOOL)canSendImagesForChat:(AIChat *)inChat
{
	return NO;
}

#pragma mark Authorization
/*!
 * @brief An authorization prompt closed, granting or denying a contact's request for authorization
 *
 * @param inWindowController The window controller which closed; an account may have kept track of what windows were showing its authorization prompts
 * @param inDict A dictionary of authorization information created by the account originally and possibly modified
 * @param authorizationResponse An AIAuthorizationResponse indicating if authorization was granted or denied or if there was no response
 */
- (void)authorizationWindowController:(NSWindowController *)inWindowController authorizationWithDict:(NSDictionary *)infoDict response:(AIAuthorizationResponse)authorizationResponse;
{}

#pragma mark Chat Commands
-(NSMenu*)actionsForChat:(AIChat*)chat
{
	return nil;
}

-(void)verifyCommand:(NSString*)commandName forChat:(AIChat*)chat
{
#warning Huh
	[self setDelegate:[[AIChatCommandsController alloc] init]];
	[[self delegate] verifyCommand:commandName forChat:chat];
	[[self delegate] setDelegate:self];
}

-(void)executeCommandWithParameters:(NSDictionary*)parameters
{
	//add stuff
}



#pragma mark delegate accessor methods
//delegate accessor methods
-(id)delegate
{
	return delegate;
}

-(void)setDelegate:(id)newDelegate;
{
	if(delegate != newDelegate)
	{
		[delegate release];
		delegate = [newDelegate retain];
	}
}
#pragma mark AppleScript
- (NSNumber *)scriptingInternalObjectID
{
	return [NSNumber numberWithInt:[[self internalObjectID] intValue]];
}

/**
 * @brief The standard objectSpecifier for this model object.
 *
 * AIAccount is contained by AIService, using the 'accounts' key.
 * Each instance has a unique integer identifier.
 */
- (NSScriptObjectSpecifier *)objectSpecifier
{
	//get my service
	AIService *theService = [self service];
	NSScriptObjectSpecifier *containerRef = [theService objectSpecifier];

	return [[[NSUniqueIDSpecifier alloc]
			 initWithContainerClassDescription:[containerRef keyClassDescription]
			 containerSpecifier:containerRef key:@"accounts"
			 uniqueID:[self scriptingInternalObjectID]] autorelease];
}

/**
 * @brief Returns the UID of this account.
 */
- (NSString *)scriptingUID
{
	return [self UID];
}

/**
 * @brief Ensures that it's impossible to set the UID of an account.
 *
 * This makes sense for the services I'm familiar with, like AIM and GTalk. It may not make sense for other protocols.
 * However, it still doesn't seem necessary to do from code.
 */
- (void)setScriptingUID:(NSString *)n
{
	[[NSScriptCommand currentCommand] setScriptErrorNumber:errOSACantAssign];
	[[NSScriptCommand currentCommand] setScriptErrorString:@"Can't dynamically change the UID of this account."];
}

/**
 * @brief Creates a chat according to the given properties.
 * @param resolvedKeyDictionary The dictionary of arguments to the 'make' command.
 *
 * This uses my own custom make<Key>WithProperties KVC method. :)
 * The idea is that be default Cocoa-AS will try to make an object using the standard alloc/init routines
 * However, you may not want that to be the case. If an AS model object implements this method, then when its the 
 * target of a 'make' command, it will be called. The method should return a new object, already assigned to a
 * container, as AICreateCommand will not do that for you.
 */
- (id)makeChatWithProperties:(NSDictionary *)resolvedKeyDictionary
{
	NSArray *participants = [resolvedKeyDictionary objectForKey:@"withContacts"];
	if (!participants) {
		[[NSScriptCommand currentCommand] setScriptErrorNumber:errOSACantAssign];
		[[NSScriptCommand currentCommand] setScriptErrorString:@"Can't create a chat without a contact!"];
		return nil;
	}
	if (![resolvedKeyDictionary objectForKey:@"newChatWindow"] && ![resolvedKeyDictionary objectForKey:@"Location"]) {
		[[NSScriptCommand currentCommand] setScriptErrorNumber:errOSACantAssign];
		[[NSScriptCommand currentCommand] setScriptErrorString:@"Can't create a chat without specifying its containing window."];
	}
	
	if ([participants count] == 1) {
		AIListContact *contact = [[participants objectAtIndex:0] objectsByEvaluatingSpecifier];
		if (!contact) {
			[[NSScriptCommand currentCommand] setScriptErrorNumber:errOSACantAssign];
			[[NSScriptCommand currentCommand] setScriptErrorString:@"Can't find that contact!"];
		}
		AIMessageWindowController *chatWindowController = nil;
		int index = -1; //at end by default
		if ([resolvedKeyDictionary objectForKey:@"newChatWindow"]) {
			//I need to put this in a new chat window
			chatWindowController = [[[AIObject sharedAdiumInstance] interfaceController] openContainerWithID:nil name:nil];
		} else {
			//I need to figure out to which chat window the location specifier is referring.
			NSLog(@"Here is the info about the location specifier: %@",[resolvedKeyDictionary objectForKey:@"Location"]);
			NSPositionalSpecifier *location = [resolvedKeyDictionary objectForKey:@"Location"];
			AIMessageWindow *chatWindow = [location insertionContainer];
			index = [location insertionIndex];
			chatWindowController = (AIMessageWindowController *)[chatWindow windowController];
		}
		
		if (!chatWindowController) {
			[[NSScriptCommand currentCommand] setScriptErrorNumber:errOSACantAssign];
			[[NSScriptCommand currentCommand] setScriptErrorString:@"Can't create chat in that chat window."];
		}
		
		AIChat *newChat = [[[AIObject sharedAdiumInstance] chatController] chatWithContact:contact];
		NSLog(@"Making new chat %@ in chat window %@:%@",newChat,chatWindowController,[chatWindowController containerID]);
		[[[AIObject sharedAdiumInstance] interfaceController] openChat:newChat inContainerWithID:[chatWindowController containerID] atIndex:index];
		return newChat;
	} else {
		if (![[self service] canCreateGroupChats]) {
			[[NSScriptCommand currentCommand] setScriptErrorNumber:errOSACantAssign];
			[[NSScriptCommand currentCommand] setScriptErrorString:@"Can't create a group chat with this service!"];
			return nil;
		}
		NSString *name = [resolvedKeyDictionary objectForKey:@"name"];
		if (!name) {
			[[NSScriptCommand currentCommand] setScriptErrorNumber:errOSACantAssign];
			[[NSScriptCommand currentCommand] setScriptErrorString:@"Can't create a group chat without a name!"];
			return nil;
		}
		//this can take a while...
		NSMutableArray *newParticipants = [[NSMutableArray alloc] init];
		for (int i=0;i<[participants count];i++) {
			[newParticipants addObject:[[participants objectAtIndex:i] objectsByEvaluatingSpecifier]];
		}
		
		//AIChat *newChat = [[[AIObject sharedAdiumInstance] chatController] chatWithName:name identifier:nil onAccount:self chatCreationInfo:nil];
		DCJoinChatViewController *chatController = [DCJoinChatViewController joinChatView];
		[chatController doJoinChatWithName:name onAccount:self chatCreationInfo:nil invitingContacts:newParticipants withInvitationMessage:@"Hey, wanna join my chat?"];
		return [[[AIObject sharedAdiumInstance] chatController] existingChatWithName:name onAccount:self];
	}
}

/**
 * @brief Returns the current status type of this account.
 */
- (AIStatusTypeApplescript)scriptingStatusType
{
	return [[self statusState] statusTypeApplescript];
}

/**
 * @brief Sets the type of the current status.
 *
 * If the current status is a temporary status, then we simply set it.
 * Otherwise, we create a temporary copy of it, and set that.
 */
- (void)setScriptingStatusType:(AIStatusTypeApplescript)scriptingType
{
	AIStatusType type;
	switch (scriptingType) {
		case AIAvailableStatusTypeAS:
			type = AIAvailableStatusType;
			break;
		case AIAwayStatusTypeAS:
			type = AIAwayStatusType;
			break;
		case AIInvisibleStatusTypeAS:
			type = AIInvisibleStatusType;
			break;
		case AIOfflineStatusTypeAS:
		default:
			type = AIOfflineStatusType;
			break;
	}
	
	AIStatus *currentStatus = [self statusState];
	if ([currentStatus mutabilityType] == AILockedStatusState || [currentStatus mutabilityType] == AISecondaryLockedStatusState) {
		switch (type) {
			case AIAvailableStatusType:
				currentStatus = [[adium statusController] availableStatus];
				break;
			case AIAwayStatusType:
				currentStatus = [[adium statusController] awayStatus];
				break;
			case AIInvisibleStatusType:
				currentStatus = [[adium statusController] invisibleStatus];
				break;
			case AIOfflineStatusType:
				currentStatus = [[adium statusController] offlineStatus];
				break;
		}
	} else {
		if ([currentStatus mutabilityType] != AITemporaryEditableStatusState) {
			currentStatus = [currentStatus mutableCopy];
			[currentStatus setMutabilityType:AITemporaryEditableStatusState];
		}
		[currentStatus setStatusType:type];
		[currentStatus setStatusName:[[[AIObject sharedAdiumInstance] statusController] defaultStatusNameForType:type]];
	}
	[[adium statusController] setActiveStatusState:currentStatus forAccount:self];
}

/**
 * @brief Returns the current status message as rich text
 */
- (NSTextStorage *)scriptingStatusMessage
{
	return [[[NSTextStorage alloc] initWithAttributedString:[[self statusState] statusMessage]] autorelease];
}

/**
 * @brief Sets the current status message
 *
 * If the current status is built in, we create a temporary copy of the current status and set that.
 */
- (void)setScriptingStatusMessage:(NSTextStorage *)message
{
	AIStatus *currentStatus = [self statusState];
	if ([currentStatus mutabilityType] != AITemporaryEditableStatusState) {
		currentStatus = [currentStatus mutableCopy];
		[currentStatus setMutabilityType:AITemporaryEditableStatusState];
	}
	if ([message isKindOfClass:[NSAttributedString class]])
		[currentStatus setStatusMessage:message];
	else
		[currentStatus setStatusMessageString:(NSString *)message];
	[[adium statusController] setActiveStatusState:currentStatus forAccount:self];
}

- (void)setScriptingStatusMessageFromScriptCommand:(NSScriptCommand *)c
{
	NSAttributedString	*messageString = [[c evaluatedArguments] objectForKey:@"WithMessage"];
	if (messageString)
		[self setScriptingStatusMessage:messageString];	
}

/**
 * @brief Tells this account to be available, with an optional temporary status message.
 */
- (void)scriptingGoAvailable:(NSScriptCommand *)c
{
	[[adium statusController] setActiveStatusState:[[adium statusController] availableStatus] forAccount:self];
	
	[self setScriptingStatusMessageFromScriptCommand:c];
}

/**
 * @brief Tells this account to be online, with an optional temporary status message.
 */
- (void)scriptingGoOnline:(NSScriptCommand *)c
{
	if ([self statusType] == AIInvisibleStatusType) {
		[self scriptingGoAvailable:c];

	} else {		
		[self setShouldBeOnline:YES];
		
		[self setScriptingStatusMessageFromScriptCommand:c];
	}
}

/**
 * @brief Tells this account to be offline, with an optional temporary status message.
 */
- (void)scriptingGoOffline:(NSScriptCommand *)c
{
	[self setShouldBeOnline:NO];

	[self setScriptingStatusMessageFromScriptCommand:c];
}

/**
 * @brief Tells this account to be away, with an optional temporary status message.
 */
- (void)scriptingGoAway:(NSScriptCommand *)c
{
	[[adium statusController] setActiveStatusState:[[adium statusController] awayStatus] forAccount:self];

	[self setScriptingStatusMessageFromScriptCommand:c];
}

/**
 * @brief Tells this account to be invisible.
 */
- (void)scriptingGoInvisible:(NSScriptCommand *)c
{
	[[adium statusController] setActiveStatusState:[[adium statusController] invisibleStatus] forAccount:self];
	
	[self setScriptingStatusMessageFromScriptCommand:c];
}

@end
